/* 
   Out Run memory test. Scavenged startup code and helper macros from the bootloader.
   Written in assembler in order to avoid stack usage during the memory test of the main RAMs.

   Outrun/68000 assembler version (c) 2017 Alex Bartholomeus.   
   For the latest version, visit: http://www.aaldert.com/outrun/memtest.html
*/   

.include "config.inc"
.include "textutil.inc"
.include "ramless.inc"
.include "shared.inc"

.text 

.ifdef HAVE_SUBCPU

# Reset SUB cpu.
.global ResetSubCPU
ResetSubCPU:
  /* Reset second CPU, and zero out the initial program counter */
  reset
  nop
  nop
  clr.l SUBRAM_BASE+4
  clr.l SUBRAM_BASE
  nop
  nop
  reset
  RETURN

# Remote procedures.
.equ memTestDataBus8Remote, 0
.equ memTestAddressBusRemote, 1
.equ memTestDeviceRemote, 2

# Marshall procedure on second CPU. %D2 = function index.
.global ExecSubCPU
ExecSubCPU:
  move.l %A0, SUB_Function_A0
  move.l %D0, SUB_Function_D0
  move.l %D1, SUB_Function_D1
  move.l %D2, SUB_Function
  move.l #1,  SUB_Function_Busy
  move.l #MAGIC_RPC_START, SUB_Magic
  nop
  nop
  nop
  nop
  nop
  nop
_wait_subcpu_loop:
  tst.l SUB_Function_Busy
  bne _wait_subcpu_loop

  move.l SUB_Function_A0, %A0
  move.l SUB_Function_D0, %D0
  move.l SUB_Function_D1, %D1
  
  RETURN
  
.endif /* HAVE_SUBCPU */

.macro WAIT_VBLANK
  move.b #1, _vblank_wait_flag
1:
  tst.b _vblank_wait_flag
  bne 1b
.endm

# Assumes IRQ4 is enabled.  
.global SwapRoad
SwapRoad:
  WAIT_VBLANK
  tst.b ROADCTRL_BASE+1 /* Swap from main cpu; Works in MAME. Not sure about real HW. */
  WAIT_VBLANK  
  RETURN
  
# Initialize/restore palette.

.global InitPalette
InitPalette:
  /* Set road palette to background color */
  move.w #0x0842, %D0

.ifdef HAVE_SUBCPU
  cmp.l #0xEFC2014, _magic_value /* bootloader */
  beq _bgcolor_set
  
# move.w #0x0428, %D0 /* "Oud-roze" */
  move.w #0x0824, %D0 /* Purplish blue */
  cmp.l #0xEFC2017, _magic_value /* memtest */
  beq _bgcolor_set
  
  # Nothing interesting found.
  move.w #0x0666, %D0 /* Pick a dreary color here */
.endif /* HAVE_SUBCPU */
  
_bgcolor_set:
  move.l #(PALETTE_BASE+(0x400*2)), %A0
  move.l #(PALETTE_BASE+(0x420*2)), %A1
  move.l #(PALETTE_BASE+(0x780*2)), %A2
  move.w #0xf, %D1
_roadpalette_loop:
  move.w %D0, (%A0)+ /* 16 entries */
  move.w %D0, (%A1)+ /* 32 entries */
  move.w %D0, (%A1)+
  move.w %D0, (%A2)+ /* 128 entries */
  move.w %D0, (%A2)+
  move.w %D0, (%A2)+
  move.w %D0, (%A2)+
  move.w %D0, (%A2)+
  move.w %D0, (%A2)+
  move.w %D0, (%A2)+
  move.w %D0, (%A2)+
  dbra   %D1, _roadpalette_loop

  /* Init text palette */
  move.l #_textpalette_end, %D0
  sub.l #_textpalette, %D0
  lsr.l #1, %D0
  subq.l #1, %D0 /* Number of colors - 1 */  
  move.l #_textpalette, %A0
  move.l #PALETTE_BASE, %A1
  
_textpalette_loop:
  move.w (%A0)+, (%A1)+
  dbra %D0, _textpalette_loop
  
  RETURN

# Init/restore tile map.
.global InitTile
InitTile:
  /* Set and clear one tile map (#0) */
  clr.w TILE_REGISTERS
  clr.w TILE_REGISTERS+2
  move.w #0x3ff, %D0
  move.l #0x00200020, %D1
  move.l #TILERAM_BASE, %A0
_tileclear_loop:
  move.l %D1, (%A0)+
  dbra %D0, _tileclear_loop

  WATCHDOG_CLEAR
  
  /* Reset all tile registers; this also sets the tile maps to use map #0 (which we just cleared) */
  move.w #0x5e, %D0
  move.l #TILE_REGISTERS, %A0
 _tilereg_loop:
  clr.l (%A0)+
  dbra %D0, _tilereg_loop

  RETURN

# Save text contents to tile memory.
.global SaveText
SaveText:
  move.w #((TEXT_WIDTH*TEXT_HEIGHT)-1), %D0
  move.l #TEXTRAM_BASE, %A0
  move.l #TILERAM_BASE, %A1 /* We could actually just use an invisible page for this */
_savetext_loop:
  move.w (%A0)+, (%A1)+
  dbra %D0, _savetext_loop
  sub.l #TEXTRAM_BASE, %A4
  add.l #TILERAM_BASE, %A4
  RETURN
  
# Restore text contents to text memory.
.global RestoreText
RestoreText:
  move.w #((TEXT_WIDTH*TEXT_HEIGHT)-1), %D0
  move.l #TILERAM_BASE, %A0
  move.l #TEXTRAM_BASE, %A1
_restoretext_loop:
  move.w (%A0)+, (%A1)+
  dbra %D0, _restoretext_loop
  sub.l #TILERAM_BASE, %A4
  add.l #TEXTRAM_BASE, %A4

  WATCHDOG_CLEAR
  jmp InitTile /* Also clear out tile ram now */
  
/*
   Entry point.
*/
.text
.global _start
_start:
  /* Assume interrupts disabled and in supervisor mode */
  
  /* This initializes the memory controller - 16 registers (size+flags?, base address) */
  lea    0xffffff20.l, %A0
  lea    memory_map_5195, %A1
  move.l #0xf, %D0
memory_mapper_loop:
  move.w (%A1)+, (%A0)+
  dbra   %D0, memory_mapper_loop

.ifdef HAVE_SUBCPU
  CALL ResetSubCPU
.endif

  WATCHDOG_CLEAR
  
  /* Init PPI, No audio and no screen. */
  move.b #0x90, IO_PPI_CONTROL /* Control register; A=in; B/C=out */
  clr.b  IO_PPI_PORT_C         /* Clear port C, resetting audio and blanking the screen */

  move.b #0x81, IO_DIGITAL_OUT /* Default value: 0x81; Internal and external mute */

  /* Clear sprites, put an empty entry at the start of the list and flip */
  move.l #0xc0000000, SPRITE_BASE
  clr.l (SPRITE_BASE+0x4)
  clr.l (SPRITE_BASE+0x8)
  clr.l (SPRITE_BASE+0xc)
  move.b #0xff, 0x140070

  CALL InitPalette

  /* Clear text layer */
  move.l #0x00200020, %D0
  move.w #0x37f, %D1
  move.l #TEXTRAM_BASE, %A0
_textclear_loop:
  move.l %D0, (%A0)+
  dbra   %D1, _textclear_loop

  WATCHDOG_CLEAR

  CALL InitTile
  
  WATCHDOG_CLEAR

  /* Enable the screen */
  move.b #0x20, IO_PPI_PORT_C

  /* Print header */  
  # Text starting position. We'll keep that in A4 !!
  move.l #(TEXTRAM_BASE+((TEXT_INDENT+(1*TEXT_WIDTH))*2)), %A4
  PRINT _str_titlemessage, COLOR_CYAN
  NEWLINE 
  NEWLINE

  WATCHDOG_CLEAR

  /* Here we have the main loop of testing memory.
     We'll be going through a list of RAM chips to test and print the results.
	 After each test we can have a custom callback in order to restore anything we messed up during the test. */
  
  # Store pointer to current entry of our RAM list.
  .global ramInfo
  move.l #ramInfo, %A5
  
  # From here on, plz do not trash either A5 (current test) or A4 (text cursor)
  
_memtest_loop:
  
  tst.l (%A5)
  beq _memtest_done /* null pointer test; if the name is nullptr, we're done testing */

  /* Print the IC name */  
  move.l (%A5)+, %A0
  move.w #COLOR_YELLOW, %D0
  CALL(_print)
  
  /* Test data bus */
  PRINT _str_db, COLOR_WHITE
  move.l (%A5)+, %A0 /* Param for memTestDataBus */
  # a5 = +8
  add.l #10, %A5

.ifdef HAVE_SUBCPU
  btst.b #7, (%A5)
  beq _memtest_db_local
  move.l #memTestDataBus8Remote, %D2
  CALL ExecSubCPU
  bra _memtest_db_done
.endif

_memtest_db_local:
  CALL memTestDataBus8Asm /* Result in D0; nonzero means failure */
_memtest_db_done:

  sub.l #18, %A5
  tst.b %D0
  beq _memtest_db_ok
  
  /* Data bus failed */
  move.b %D0, %D1
  PRINT _str_error, COLOR_RED
  move.b %D1, %D0
  CALL _printhex2
  jmp _memtest_fail
  
_memtest_db_ok:
  WATCHDOG_CLEAR
  PRINT _str_ok, COLOR_GREEN  
  PRINT _str_ab, COLOR_WHITE
  
  /* Test address bus */
  addq.l #4, %A5
  move.l (%A5), %A0 /* Address */
  addq.l #6, %A5
  move.l (%A5), %D0 /* Size */
  move.l #2, %D1    /* Offset = 2 */
  add.l #8, %A5

.ifdef HAVE_SUBCPU  
  btst.b #7, (%A5)
  beq _memtest_ab_local
  move.l #memTestAddressBusRemote, %D2
  CALL ExecSubCPU
  bra _memtest_ab_done
.endif  

_memtest_ab_local:
  CALL memTestAddressBusAsm
_memtest_ab_done:

  sub.l #18, %A5

  move.l %A0, %D0
  tst.l %D0
  beq _memtest_ab_ok

  /* Address bus failed */
_memtest_fail_a0: /* reused for device */
  move.l %D0, %D1
  PRINT _str_error, COLOR_RED
  move.l %D1, %D0
  CALL _printhex6

_memtest_fail:
  move.b #0x9a, IO_DIGITAL_OUT /* Brake lamp */
  move.l #1000, %D0
  CALL(_sleep);
  
  jmp _memtest_next
  
_memtest_ab_ok:
  WATCHDOG_CLEAR
  PRINT _str_ok, COLOR_GREEN  
  PRINT _str_dev, COLOR_WHITE
  
  /* Full region test. This will result in garbage on screen for some tests. */
  addq.l #4, %A5
  move.l (%A5), %A0 /* Address */
  addq.l #6, %A5
  move.l (%A5), %D0 /* Size */
  subq.l #2, %A5 
  clr.l %D1
  move.w (%A5), %D1 /* Interleave */
  add.l #10, %A5

.ifdef HAVE_SUBCPU  
  btst.b #7, (%A5)
  beq _memtest_dev_local
  move.l #memTestDeviceRemote, %D2
  CALL ExecSubCPU
  bra _memtest_dev_done
.endif

_memtest_dev_local:
  CALL memTestDeviceAsm
_memtest_dev_done:
  
  sub.l #18, %A5
  move.l %A0, %D0
  tst.l %D0
  beq _memtest_dev_ok
  jmp _memtest_fail_a0
  
_memtest_dev_ok:
  PRINT _str_ok, COLOR_GREEN  
  WATCHDOG_CLEAR

  /* IC is entirely good! */
  move.b #0x9c, IO_DIGITAL_OUT /* Start lamp */
  move.l #500, %D0
  CALL(_sleep);

.global _memtest_next  
_memtest_next:  

  # Optional restoration callback.
  add.l #14, %A5
  tst.l (%A5)
  beq _memtest_nocb
  
  # Remote callback?
  addq.l #4, %A5
.ifdef HAVE_SUBCPU
  btst #7, (%A5)
  beq _memtest_localcb /* No, I like local tamale. */

  # Remote callback.
  subq.l #4, %A5 /* -4 */
  move.l (%A5), %D2
  CALL ExecSubCPU
  bra _memtest_nocb
.endif

_memtest_localcb:
  subq.l #4, %A5 /* -4 */
  move.l #_memtest_nocb, %A6
  move.l (%A5), %A0
  jmp (%A0) /* needed for indirect jump */
_memtest_nocb:
  sub.l #14, %A5 /* Already -4 */

  # Wait another 300ms after the lights go out, that way we can always the results apart.
  move.b #0x98, IO_DIGITAL_OUT /* Reset */
  move.l #300, %D0
  CALL _sleep
  
  /* Next test */
  NEWLINE
  add.l #20, %A5  /* add sizeof(struct RAMREGION) */
  jmp _memtest_loop
    
_memtest_done:
  
  # Rebooting in 9... and so on.

  NEWLINE
  PRINT _str_reboot, COLOR_RED
  subq.l #8, %A4      /* Set cursor pos right on the '9' */

  move.w #8, %D2
_restart_countdown:  
  move.l #1000, %D0
  CALL(_sleep);       /* You should think about it, take a second. */
  subq.w #1, (%A4)
  dbra %D2, _restart_countdown
  
  # Done waiting, reboot.
  
/*
  Reboots to the image in ROM.
*/
_rebootrom:

  move.w #0x2700, %SR /* Disable interrupts */
.ifdef HAVE_SUBCPU
  CALL ResetSubCPU
.endif

  move.w #0x003, %D0 /* Screen to red */
  CALL _bgcolor_set

  /* Sleep 500ms, but without resetting the watchdog. On a real board, with the watchdog in place, it will time out way faster than this. */
  move.w #500, %D0
_sleep_reboot:
  move.w #833, %D1 /* dbra is ~12 cycles, so this should be about one millisecond */
_sleep_1ms_reboot:
  dbra %D1, _sleep_1ms_reboot
  dbra %D0, _sleep_reboot
  
  /* The old fashioned way */
  jmp _start;
  
/* 
   Sleep function; D0 = amount of milliseconds (approximately); A6 = return address.
   Modifies D0 and D1
*/
_sleep:
  move.w #833, %D1 /* dbra is ~12 cycles, so this should be about one millisecond */
_sleep_one_ms:
  dbra %D1, _sleep_one_ms
  WATCHDOG_CLEAR
  dbra %D0, _sleep
  RETURN

# Dummy IRQ routine.
.global _dummyirq
_dummyirq:
  rte

# Vertical blank handler.
# This will do an auto watchdog clear; and clear a timer.
# It's intended to take at least the entire scanline.
.global _vblankirq
_vblankirq:
  clr.b _vblank_wait_flag

.ifdef CPU0
  WATCHDOG_CLEAR
.endif

  /* Wait out the rest of the scanline so the interrupt doesn't trigger twice.
     we have 262 scanlines @ 60fps = 15720 scanlines per second. So one scanline is 636.1 cycles @ 10Mhz.
     Since dbra takes at least 10 cycles, we can loop 64x to waste at least 640 cycles */
  move.l   %D0, -(%SP)
  move.w   #((CYCLES_PER_SCANLINE+9)/10), %D0
_irq4_wait:
  dbra     %D0, _irq4_wait
  /* We are now at scanline 224 (invisible) */
  move.l   (%SP)+, %D0

  rte

.global EnableIRQ4
EnableIRQ4:
  move.l #_stack_super, %A7 /* The supervisor stack pointer needs to be valid! Assumes we are still in supervisor mode. */
  move.w #0x2300, %SR
  RETURN
  
# Miscellaneous data, written to ROM.
.section .rodata

_str_reboot:
.asciz "RESTARTING IN 9..."
_str_error:
.asciz " ERR "
_str_ok:
.asciz " OK"
_str_ab:
.asciz " AB"
_str_db:
.asciz " DB"
_str_dev:
.asciz " DEV"

.section .data

/* IRQ timer ack */
_vblank_wait_flag: 
.skip 1 
